/*  
 *  libmenuthing v0.0.1
 *  simple menu library.. thing
 *  ARM (GP2X)
 *  (C) Alvin Sipraga <alvin.sipraga@gmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "libmenuthing.h"

void mt_init()
{
    /* if SDL_ttf hadsn't been initialised */
    if(!TTF_WasInit())
        TTF_Init();
}

/* add item to list */
void mt_add_item(list<mt_list> *menulist, string label,
        void (*func)(void *), void *arg)
{
    list<mt_list>::iterator iter;
    
    mt_list push_in;
    push_in.label = label;
    push_in.func = func;
    push_in.farg = arg;
//    push_in.arg1=arg1;
    menulist->push_back(push_in);
    
    iter = menulist->end();
    iter--;
    printf("added: \"%s\" to list\n", iter->label.c_str());
}


/* start menu loop */
void mt_start_menu(list<mt_list> menulist, string title, string font,
        string font_bold, SDL_Surface *screen)
{
    list<mt_list>::iterator iter;
    SDL_Rect dest;
    SDL_Surface *text;
    SDL_Color color = {255, 255, 255};
    SDL_Color selected = {255, 255, 85};
    
    /* load fonts */
    TTF_Font *font_list_normal = TTF_OpenFont(font.c_str(), 12);
    if(!font_list_normal) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }
    
    TTF_Font *font_list_selected = TTF_OpenFont(font_bold.c_str(), 12);
    if(!font_list_selected) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }
    
    TTF_Font *font_title = TTF_OpenFont(font_bold.c_str(), 16);
    if(!font_title) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }
        SDL_FillRect(screen,NULL,SDL_MapRGB(screen->format,0,0,0));
	    SDL_Flip(screen);
    dest.x = 20;
    dest.y = 20;
    
    /* render title */
    text = TTF_RenderText_Solid(font_title, title.c_str(), color);
    
    /* blit it to surface */
    if(SDL_BlitSurface(text, NULL, screen, &dest))
        printf("%s", SDL_GetError());
    
    /* loop through the list */
    for(iter = menulist.begin(), dest.x = 40, dest.y = 60;
            iter != menulist.end(); iter++, dest.y+=15) {
        /* if it's the first one */
        if(iter == menulist.begin()) {
            /* render font */
            text = TTF_RenderText_Solid(font_list_selected,
                    iter->label.c_str(), selected);
            
            /* and blit */
            if(SDL_BlitSurface(text, NULL, screen, &dest))
                printf("%s", SDL_GetError());
            
            /* set place */
            iter->dest.x = dest.x;
            iter->dest.y = dest.y;
            
            /* first item starts selected */
            iter->selected = 1;
        }
        else {
            /* render */
            text = TTF_RenderText_Solid(font_list_normal, iter->label.c_str(),
                    color);
            
            /* blit */
            if(SDL_BlitSurface(text, NULL, screen, &dest))
                printf("%s", SDL_GetError());
            
            /* set place */
            iter->dest.x = dest.x;
            iter->dest.y = dest.y;
            
            /* first item starts selected */
            iter->selected = 0;
        }
    }
    
    /* flip screen */
    SDL_Flip(screen);
    
    /* event block */
    SDL_Event event;
    
    /* pend events, will block until item is selected */
    while(SDL_WaitEvent(&event) != 0) {
        switch(event.type) {
            /* we're only interested in keys */
            case SDL_JOYBUTTONDOWN:
                /* if it's down */
                if(event.jbutton.button == MT_ACT_DOWN) {
                    /* find the selected item */
                    for(iter = menulist.begin(); ; iter++)
                        if(iter->selected == 1)
                            break;
                    
                    /* set mask colour */
                    SDL_Color black = {0, 0, 0};
                    
                    /* render mask */
                    text = mt_rend_mask(iter->label, black, font, font_bold);
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit mask */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* render new text */
                    if(!(text = TTF_RenderText_Solid(font_list_normal,
                            iter->label.c_str(), color)))
                        printf("%s", TTF_GetError());
                    
                    /* blit new text */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* no longer selected */
                    iter->selected = 0;
                    
                    /* iterate through 1 */
                    iter++;
                    
                    /* if it's at the end, make it go to the beginning */
                    if(iter == menulist.end())
                        iter = menulist.begin();
                        
                    /* render mask */
                    text = mt_rend_mask(iter->label, black, font, font_bold);
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit mask */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* render */
                    if(!(text = TTF_RenderText_Solid(font_list_selected,
                            iter->label.c_str(), selected)))
                        printf("%s", TTF_GetError());
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* set as selected */
                    iter->selected = 1;
                    
                    /* flip */
                    if(SDL_Flip(screen))
                        printf("%s", SDL_GetError());
                }
                else if(event.jbutton.button == MT_ACT_UP) {
                    /* find selected item */
                    for(iter = menulist.begin(); ; iter++)
                        if(iter->selected == 1)
                            break;
                    
                    /* set mask colour */
                    SDL_Color black = {0, 0, 0};
                    
                    /* render mask */
                    if(!(text = TTF_RenderText_Solid(font_list_selected,
                            iter->label.c_str(), black)))
                        printf("%s", TTF_GetError());
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit mask */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* render new text */
                    if(!(text = TTF_RenderText_Solid(font_list_normal,
                            iter->label.c_str(), color)))
                        printf("%s", TTF_GetError());
                            
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit it */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* no longer selected */
                    iter->selected = 0;
                    
                    /* if it's at the top, go to the bottom */
                    if(iter == menulist.begin())
                        iter = menulist.end();
                    
                    iter--;
                    
                    /* render mask */
                    text = mt_rend_mask(iter->label, black, font, font_bold);
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit mask */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* render text */
                    if(!(text = TTF_RenderText_Solid(font_list_selected,
                            iter->label.c_str(), selected)))
                        printf("%s", TTF_GetError());
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit text */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    
                    /* set as selected */
                    iter->selected = 1;
                    
                    /* flip */
                    if(SDL_Flip(screen))
                        printf("%s", SDL_GetError());
                }
                /* else if it's a selection */
                else if(event.jbutton.button == MT_ACT_SELECT) {
                    SDL_Color action = {170, 0, 0};
                    /* go to selected item */
                    for(iter = menulist.begin(); ; iter++)
                        if(iter->selected == 1)
                            break;
                    
                    /* render red text */
                    if(!(text = TTF_RenderText_Solid(font_list_selected,
                            iter->label.c_str(), action)))
                        printf("%s", TTF_GetError());
                    
                    dest.x = iter->dest.x;
                    dest.y = iter->dest.y;
                    
                    /* blit text */
                    if(SDL_BlitSurface(text, NULL, screen, &dest))
                        printf("%s", SDL_GetError());
                    SDL_Flip(screen);
                    
                    SDL_Event subevent;
                    
                    while(SDL_WaitEvent(&subevent) != 0) {
                        if(subevent.type == SDL_JOYBUTTONUP &&
                                event.jbutton.button == MT_ACT_SELECT) {
                            /* render selected text */
                            if(!(text = TTF_RenderText_Solid(font_list_selected,
                                    iter->label.c_str(), selected)))
                                printf("%s", TTF_GetError());
                            
                            dest.x = iter->dest.x;
                            dest.y = iter->dest.y;
                            
                            /* blit text */
                            if(SDL_BlitSurface(text, NULL, screen, &dest))
                                printf("%s", SDL_GetError());
                            SDL_Flip(screen);
                            
                            /* run the function if specified when added */
                            if(iter->func != NULL) {
                                iter->func(iter->farg);
    SDL_FillRect(screen,NULL,SDL_MapRGB(screen->format,0,0,0));
        SDL_Flip(screen);
//				return;
                            }
//                            return;
                            break;
                        }
                    }
                }
                else if(event.jbutton.button == MT_ACT_QUIT) {
                    return;
                }
        }
    }
}

SDL_Surface *mt_rend_mask(string label, SDL_Color color,
        string font, string font_bold)
{
    SDL_Surface *text;
    
    /* load fonts */
    TTF_Font *font_list_normal = TTF_OpenFont(font.c_str(), 12);
    if(!font_list_normal) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }
    
    TTF_Font *font_list_selected = TTF_OpenFont(font_bold.c_str(), 12);
    if(!font_list_selected) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        exit(1);
    }
    
    /* render mask */
    if(!(text = TTF_RenderText_Solid(font_list_selected,
            label.c_str(), color)))
        printf("%s", TTF_GetError());
    
    if(!(SDL_BlitSurface(TTF_RenderText_Solid(font_list_normal,
            label.c_str(), color), NULL, text, NULL)))
        printf("%s", TTF_GetError());
    
    return text;
}
